# -*- coding: utf-8 -*-


class CgroupsException(Exception):
    pass


class CgroupSubsysFormatInvalidError(CgroupsException):
    def __init__(self, line, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.line = line

    def __str__(self):
        return "invalid format for CGroupSubsys: %s" % self.line


class MountPointFormatInvalidError(CgroupsException):
    def __init__(self, line, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.line = line

    def __str__(self):
        return "invalid format for MountPoint: %s" % self.line


class PathNotExposedFromMountPointError(CgroupsException):
    def __init__(self, mount_point, root, path, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.mount_point = mount_point
        self.root = root
        self.path = path

    def __str__(self):
        return "path %s is not a descendant of mount point root %s and cannot be exposed from %s" % (
            self.path, self.root, self.mount_point)



# -*- coding: utf-8 -*-

from collections import UserDict
from typing import Dict
from .subsys import parse_cgroup_subsystems
from .mountpoint import MountPoint, parse_mount_info
from .errors import (
    CgroupSubsysFormatInvalidError,
    MountPointFormatInvalidError,
    PathNotExposedFromMountPointError)
from .cgroup import CGroup


# _cgroupFSType is the Linux CGroup file system type used in
# `/proc/$PID/mountinfo`.
_cgroupFSType = "cgroup"
# _cgroupSubsysCPU is the CPU CGroup subsystem.
_cgroupSubsysCPU = "cpu"
# _cgroupSubsysCPUAcct is the CPU accounting CGroup subsystem.
_cgroupSubsysCPUAcct = "cpuacct"
# _cgroupSubsysCPUSet is the CPUSet CGroup subsystem.
_cgroupSubsysCPUSet = "cpuset"
# _cgroupSubsysMemory is the Memory CGroup subsystem.
_cgroupSubsysMemory = "memory"

# cfs 是完全公平调度器的缩写，即按照比例进行cpu分配调度
# cpu.cfs_period_us和cpu.cfs_quota_us来限制该组中的所有进程在单位时间里可以使用的cpu时间
# 例如：周期为1秒，假设CPU>=4核心，表示这个组在一个使用周期(1s)内可以跑满4核资源
# cpu.cfs_period_us = 1000000
# cpu.cfs_quota_us = 4000000
# _cgroupCPUCFSQuotaUsParam is the file name for the CGroup CFS quota
# parameter.
# 设置cpu.cfs_period_us周期内cgroup可使用的cpu
# 默认 -1，表示使用的CPU不受cgroup限制；最小值为1ms(1000)，最大值为1s
_cgroupCPUCFSQuotaUsParam = "cpu.cfs_quota_us"
# _cgroupCPUCFSPeriodUsParam is the file name for the CGroup CFS period
# parameter.
# 周期内允许占用的CPU时间，默认为 100000，单位为微秒us
_cgroupCPUCFSPeriodUsParam = "cpu.cfs_period_us"

# /proc/self 下是当前进程的信息
# /proc/self/cgroup 是当前进程的cgroup信息
_procPathCGroup = "/proc/self/cgroup"
# 记录当前系统所有挂载文件系统的信息
_procPathMountInfo = "/proc/self/mountinfo"


class CGroups(UserDict):
    def cpu_quota(self):
        cpu_cgroup = self.get(_cgroupSubsysCPU)
        if not cpu_cgroup:
            return -1, False

        try:
            cfs_quota_us = cpu_cgroup.read_int(_cgroupCPUCFSQuotaUsParam)
        except ValueError:
            return -1, False
        if cfs_quota_us <= 0:
            return -1, False

        try:
            cfs_period_us = cpu_cgroup.read_int(_cgroupCPUCFSPeriodUsParam)
        except ValueError:
            return -1, False

        if cfs_period_us == 0:
            return -1, False

        quota = float(cfs_quota_us) / float(cfs_period_us)
        return quota, True


def new_cgroups(proc_path_mount_info,
                proc_path_cgroup: str) -> Dict[str, CGroups]:
    """

    :param proc_path_mount_info: 当前系统所有挂载文件系统的信息
    :param proc_path_cgroup: 当前进程的cgroup信息
    :return:
    """
    # 解析当前进程的cgroup子系统信息
    try:
        cgroup_subsystems = parse_cgroup_subsystems(proc_path_cgroup)
    except (ValueError, CgroupSubsysFormatInvalidError):
        return
    cgroups = {}

    def new_mount_point(mp: MountPoint) -> None:
        # 判断文件系统是否是cgroup
        if mp.fs_type != _cgroupFSType:
            return
        # 遍历超级快选项
        for opt in mp.super_options:
            # 判断cgroups子系统是否存在
            subsys = cgroup_subsystems.get(opt)
            if not subsys:
                continue
            # 获得cgroup的挂载点路径
            cgroup_path = mp.translate(subsys.name)
            cgroups[opt] = CGroup(cgroup_path)

    # 解析挂载的文件系统信息
    try:
        parse_mount_info(proc_path_mount_info, new_mount_point)
    except (FileNotFoundError, MountPointFormatInvalidError, PathNotExposedFromMountPointError):
        return
    return CGroups(cgroups)


def new_cgroups_for_current_process():
    return new_cgroups(_procPathMountInfo, _procPathCGroup)



# -*- coding: utf-8 -*-

import os
from typing import List, Callable
from .errors import (
    PathNotExposedFromMountPointError,
    MountPointFormatInvalidError)


_mountInfoSep = " "
_mountInfoOptsSep = ","
_mountInfoOptionalFieldsSep = "-"


_miFieldIDMountID = 0
_miFieldIDParentID = 1
_miFieldIDDeviceID = 2
_miFieldIDRoot = 3
_miFieldIDMountPoint = 4
_miFieldIDOptions = 5
_miFieldIDOptionalFields = 6
_miFieldCountFirstHalf = 7


_miFieldOffsetFSType = 0
_miFieldOffsetMountSource = 1
_miFieldOffsetSuperOptions = 2
_miFieldCountSecondHalf = 3

_miFieldCountMin = _miFieldCountFirstHalf + _miFieldCountSecondHalf


class MountPoint:
    """
    MountPoint is the data structure for the mount points in
    `/proc/$PID/mountinfo`. See also proc(5) for more information.
    """

    def __init__(
            self,
            mount_id: int,
            parent_id: int,
            device_id: str,
            root: str,
            mount_point: str,
            options: List[str],
            optional_fields: List[str],
            fs_type: str,
            mount_source: str,
            super_options: List[str]):
        self.mount_id = mount_id
        self.parent_id = parent_id
        self.device_id = device_id
        self.root = root
        self.mount_point = mount_point
        self.options = options
        self.optional_fields = optional_fields
        self.fs_type = fs_type
        self.mount_source = mount_source
        self.super_options = super_options

    def translate(self, abs_path: str) -> str:
        rel_path = os.path.relpath(abs_path, self.root)

        if rel_path == ".." or rel_path.startswith("../"):
            raise PathNotExposedFromMountPointError(
                self.mount_point, self.root, abs_path)

        if rel_path != ".":
            return os.path.join(self.mount_point, rel_path)
        return self.mount_point


def new_mount_point_from_line(line: str) -> MountPoint:
    """解析文件系统挂载信息

    字段详情：https://man7.org/linux/man-pages/man5/proc.5.html
    """
    # 分割每一行
    fields = line.split(_mountInfoSep)
    # 每一行有至少10个属性，可能大于10个
    if len(fields) < _miFieldCountMin:
        raise MountPointFormatInvalidError(line)
    # 第0列：mount ID,挂载点的唯一标识
    mount_id = int(fields[_miFieldIDMountID])
    # 第1列：parent ID,当前挂载点的父挂载点的ID
    parent_id = int(fields[_miFieldIDParentID])
    # 第2列：major:minor, files的st_dev的值
    # 第3列：root: 文件系统的根挂载点
    # 第4列：mount point: 相对于进程根目录的挂载点
    # 第5列：mount options: 预挂载选项
    # 第6列：options fields: tag:[value]类型的字段

    for i, field in enumerate(fields[_miFieldIDOptionalFields:]):
        # 第7列：sparator: options fields结束标志
        if field == _mountInfoOptionalFieldsSep:
            fs_type_start = _miFieldIDOptionalFields + i + 1
            if len(fields) != fs_type_start + _miFieldCountSecondHalf:
                raise MountPointFormatInvalidError(line)
            # 第8列：file systemtype: 文件系统的名称,以type[.subtype]的方式命名
            mi_field_id_fs_type = _miFieldOffsetFSType + fs_type_start
            # 第9列：mount source: 文件特定信息
            mi_field_id_mount_source = _miFieldOffsetMountSource + fs_type_start
            # 第10列：super options: 超级块选项
            mi_field_id_super_options = _miFieldOffsetSuperOptions + fs_type_start

            # 创建文件系统挂载点对象
            mount_point = MountPoint(
                mount_id=mount_id,
                parent_id=parent_id,
                device_id=fields[_miFieldIDDeviceID],
                root=fields[_miFieldIDRoot],
                mount_point=fields[_miFieldIDMountPoint],
                options=fields[_miFieldIDOptions].split(_mountInfoOptsSep),
                optional_fields=fields[_miFieldIDOptionalFields:(fs_type_start - 1)],
                fs_type=fields[mi_field_id_fs_type],
                mount_source=fields[mi_field_id_mount_source],
                super_options=fields[mi_field_id_super_options].split(_mountInfoOptsSep)
            )
            return mount_point


def parse_mount_info(proc_path_mount_info: str, new_mount_point: Callable[[MountPoint], None]):
    """解析挂载的文件系统信息

    Args:
        proc_path_mount_info(str): 当前系统所有挂载文件系统的信息
        new_mount_point: 每一个挂载的文件系统的解析回调
    """
    with open(proc_path_mount_info) as reader:
        for line in reader:
            line = line.strip()
            if not line:
                continue
            # 解析每一个挂载的文件系统的信息，返回文件系统挂载点对象
            mount_point = new_mount_point_from_line(line)
            # 执行回调
            new_mount_point(mount_point)



# -*- coding: utf-8 -*-

import os


class CGroup:
    def __init__(self, path):
        self.path = path

    def path(self):
        return self.path

    def param_path(self, param):
        return os.path.join(self.path, param)

    def read_first_line(self, param):
        file_path = self.param_path(param)
        with open(file_path, "r") as reader:
            for row in reader:
                if row:
                    return row.strip()

    def read_int(self, param):
        try:
            first_row = self.read_first_line(param)
        except FileNotFoundError:
            return 0
        try:
            return int(first_row) if first_row else 0
        except ValueError:
            return 0



# -*- coding: utf-8 -*-

import multiprocessing
from automaxprocs.runtime import cpu_quota_to_max_procs


def get_cpu_count():
    return multiprocessing.cpu_count()


def get_max_procs():
    max_procs, status = cpu_quota_to_max_procs(1)
    if max_procs == -1:
        return get_cpu_count()
    return max_procs



# -*- coding: utf-8 -*-

from __future__ import unicode_literals
from setuptools import find_packages

import logging
import os
import codecs
import sys
from io import open

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup
from setuptools.command.test import test as TestCommand
from importlib import import_module


logger = logging.getLogger(__name__)

version = import_module('automaxprocs.version').version


def read(filename):
    """Read and return `filename` in root dir of project and return string ."""
    here = os.path.abspath(os.path.dirname(__file__))
    return codecs.open(os.path.join(here, filename), 'r').read()


test_requirements = [
    'pytest',
]


class PyTest(TestCommand):
    user_options = [("pytest-args=", "a", "Arguments to pass to pytest")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.pytest_args = ""

    def run_tests(self):
        import shlex
        import pytest
        errno = pytest.main(shlex.split(self.pytest_args))
        sys.exit(errno)


readme = os.path.join(os.path.dirname(__file__), 'README.md')
long_description = open(readme, encoding='utf-8').read() + '\n\n'

install_requires = []
if os.path.exists("requirements.txt"):
    install_requires = read("requirements.txt").split()
    if "win" in sys.platform:
        for item in ['pexpect']:
            try:
                install_requires.remove(item)
            except ValueError as e:
                pass


install_requires.extend([

])


def do_setup():
    setup(
        name='pyautomaxprocs',
        version=version,
        description='Automatically get Linux container CPU quota like uber automaxprocs.',
        long_description=long_description,
        long_description_content_type='text/markdown',
        author='jinyinqiao',
        author_email='jinyinqiao@gmail.com',
        license='MIT',
        packages=find_packages(exclude=['tests*']),
        include_package_data=True,
        install_requires=install_requires,
        zip_safe=False,
        scripts=None,
        keywords='automaxprocs',
        url='https://github.com/fengzhongzhu1621/automaxprocs',
        classifiers=[
            'Intended Audience :: Developers',
            'License :: OSI Approved :: MIT License',
            'Natural Language :: English',
            'Operating System :: MacOS',
            'Operating System :: POSIX',
            'Environment :: Console',
            'Programming Language :: Python :: 3',
            'Programming Language :: Python :: 3.6',
        ],
        setup_requires=[],
        extras_require={
            'dev': ['pytest', 'bump2version'],
        },
        test_suite='tests',
        tests_require=test_requirements,
        cmdclass={'test': PyTest}
    )


if __name__ == "__main__":
    do_setup()



# -*- coding: utf-8 -*-

CPUQuotaStatus = int

# CPUQuotaUndefined is returned when CPU quota is undefined
CPUQuotaUndefined = 0
# CPUQuotaUsed is returned when a valid CPU quota can be used
CPUQuotaUsed = 1
# CPUQuotaMinUsed is return when CPU quota is smaller than the min value
CPUQuotaMinUsed = 2



# -*- coding: utf-8 -*-

import platform
import math
from .runtime import (
    CPUQuotaStatus,
    CPUQuotaUndefined,
    CPUQuotaMinUsed,
    CPUQuotaUsed)
from ..cgroups.cgroups import new_cgroups_for_current_process


def cpu_quota_to_max_procs(min_value: int) -> (int, CPUQuotaStatus):
    # 当前进程的cgroup信息
    cgroups = new_cgroups_for_current_process()
    if not cgroups:
        return -1, CPUQuotaUndefined

    quota, defined = cgroups.cpu_quota()
    if not defined:
        return -1, CPUQuotaUndefined

    max_procs = int(math.floor(quota))

    if min_value > 0 and max_procs < min_value:
        return min_value, CPUQuotaMinUsed

    return max_procs, CPUQuotaUsed



# -*- coding: utf-8 -*-

from typing import List, Dict
from .errors import CgroupSubsysFormatInvalidError


_cgroupSep = ":"
_cgroupSubsysSep = ","

_csFieldIDID = 0
_csFieldIDSubsystems = 1
_csFieldIDName = 2
_csFieldCount = 3


class CGroupSubsys:
    def __init__(self, subsys_id: int, subsystems: List[str], name: str):
        self.id = subsys_id
        self.subsystems = subsystems
        self.name = name


def new_cgroup_subsys_from_line(line: str) -> CGroupSubsys:
    fields = line.split(_cgroupSep)
    if len(fields) != _csFieldCount:
        raise CgroupSubsysFormatInvalidError(line)

    # 第0列是子系统ID
    subsys_id = int(fields[_csFieldIDID])
    # 可能有多个子系统
    subsystems = fields[_csFieldIDSubsystems].split(_cgroupSubsysSep)
    # 子系统路径
    name = fields[_csFieldIDName]
    subsys = CGroupSubsys(subsys_id, subsystems, name)
    return subsys


def parse_cgroup_subsystems(proc_path_cgroup: str) -> Dict[str, CGroupSubsys]:
    """解析当前进程的cgroup子系统信息 ."""
    subsystems = {}
    with open(proc_path_cgroup, mode='r') as reader:
        for line in reader:
            line = line.strip()
            if not line:
                continue
            cgroup = new_cgroup_subsys_from_line(line)
            for subsys in cgroup.subsystems:
                subsystems[subsys] = cgroup

    return subsystems
